/* A simple simulator for the osu-8 instructional microprocessor */

#include <stdio.h>
#include <time.h>

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC 1000000
#endif

extern char * disasm(int byte, int memory_location);
extern void hexout(int byte, int memory_location, int end);

int	a, b, p1, p2, sp, pc;	/* the registers, global variables */
int	c, v, z, p;		/* the status bits */
int	memory[65536];		/* the memory is also global */
int	in[2], out[4];		/* I/O bits */

int	usage_stat[256];	/* number of times each instruction executed */
int	jcu_t=0, jcu_nt=0, jcd_t=0, jcd_nt=0;	/* count branch results also */
int	debug = 0;

FILE	*fhex;

void reg()  /* print out a summary of the registers */
{
	printf("Reg: A:%02X B:%02X  P1:%04X P2:%04X", a, b, p1, p2);
	printf("  C:%d V:%d Z:%d P:%d  SP:%04X PC:%04X\n", c, v, z, p, sp, pc);
}

void one_instruction()  /* simulate the current instruction */
{
	register int	opcode, operand;
	int		alu;

	opcode = (memory[pc] >> 4) & 15;  /* fetch the instruction */
	operand = memory[pc] & 15;
	usage_stat[memory[pc] & 255]++;
	pc = (pc + 1) & 65535;
	switch (opcode) {
	case (2):  /* jump up if carry set */
		if (debug) printf("JCU %04X", (pc - (operand << 2)) & 65532);
		if (c) {
			pc = (pc - ((operand+1) << 2)) & 65532;
			jcu_t++;
		} else jcu_nt++;
		break;
	case (3):  /* jump down if carry set */
		if (debug) printf("JCD %04X", (pc + (operand << 2)) & 65532);
		if (c) {
			pc = (pc + (operand << 2)) & 65532;
			jcd_t++;
		} else jcd_nt++;
		break;
	case (4):  /* push, pop, jump, call, ret, reti, nop */
		switch (operand) {
		case 0:
			memory[sp] = a;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH A");
			break;
		case 1:
			sp = (sp - 1) & 65535;
			a = memory[sp] & 255;
			if (debug) printf("POP A");
			break;
		case 2:
			memory[sp] = b;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH B");
			break;
		case 3:
			sp = (sp - 1) & 65535;
			b = memory[sp] & 255;
			if (debug) printf("POP B");
			break;
		case 4:
			memory[sp] = p1 & 255;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH P1L");
			break;
		case 5:
			p1 &= 65280;
			sp = (sp - 1) & 65535;
			p1 |= memory[sp] & 255;
			if (debug) printf("POP P1L");
			break;
		case 6:
			memory[sp] = (p1 >> 8) & 255;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH P1H");
			break;
		case 7:
			p1 &= 255;
			sp = (sp - 1) & 65535;
			p1 |= (memory[sp] & 255) << 8;
			if (debug) printf("POP P1H");
			break;
		case 8:
			memory[sp] = p2 & 255;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH P2L");
			break;
		case 9:
			p2 &= 65280;
			sp = (sp - 1) & 65535;
			p2 |= memory[sp] & 255;
			if (debug) printf("POP P2L");
			break;
		case 10:
			memory[sp] = (p2 >> 8) & 255;
			sp = (sp + 1) & 65535;
			if (debug) printf("PUSH P2H");
			break;
		case 11:
			p2 &= 255;
			sp = (sp - 1) & 65535;
			p2 |= (memory[sp] & 255) << 8;
			if (debug) printf("POP P2H");
			break;
		case 12:
			if (debug) printf("CALL @P1");
			memory[sp] = (pc >> 8) & 255;
			sp = (sp + 1) & 65535;
			memory[sp] = pc & 255;
			sp = (sp + 1) & 65535;
			pc = p1;
			break;
		case 13:
			if (debug) printf("JUMP @P1");
			pc = p1;
			break;
		case 14:
		case 15:
			if (debug) printf("RET");
			sp = (sp - 1) & 65535;
			pc = (memory[sp] & 255);
			sp = (sp - 1) & 65535;
			pc |= (memory[sp] & 255) << 8;
			break;
		}
		break;
	case (5):  /* load, store */
		switch (operand) {
		case 0:
			a = memory[p1] & 255;
			if (debug) printf("LOAD @P1, A");
			break;
		case 1:
			a = memory[p2] & 255;
			if (debug) printf("LOAD @P2, A");
			break;
		case 2:
			b = memory[p1] & 255;
			if (debug) printf("LOAD @P1, B");
			break;
		case 3:
			b = memory[p2] & 255;
			if (debug) printf("LOAD @P2, B");
			break;
		case 8:
			memory[p1] = a;
			if (debug) printf("STORE A, @P1");
			break;
		case 9:
			memory[p2] = a;
			if (debug) printf("STORE A, @P2");
			break;
		case 10:
			memory[p1] = b;
			if (debug) printf("STORE B, @P1");
			break;
		case 11:
			memory[p2] = b;
			if (debug) printf("STORE B, @P2");
			break;
		case 4:
		case 5:
		case 6:
		case 7:
		case 12:
		case 13:
		case 14:
		case 15:
			if (debug) printf("<undefined opcode>");
			/* these 8 instrustions are undefined */
		}
		break;
	case (6):  /* move, clrb/setb/cplb */
		switch (operand) {
		case 0:
			if (debug) printf("MOVE IN0, C");
			c = in[0];
			break;
		case 1:
			if (debug) printf("MOVE IN1, C");
			c = in[1];
			break;
		case 2:
			if (debug) printf("MOVE V, C");
			c = v;
			break;
		case 3:
			if (debug) printf("MOVE Z, C");
			c = z;
			break;
		case 4:
			if (debug) printf("MOVE P, C");
			c = p;
			break;
		case 5:
			if (debug) printf("CLR C");
			c = 0;
			break;
		case 6:
			if (debug) printf("SET C");
			c = 1;
			break;
		case 7:
			if (debug) printf("CPL C");
			if (c) c = 0; else c = 1;
			break;
		case 8:
			if (debug) printf("MOVE C, OUT0");
			out[0] = c;
			break;
		case 9:
			if (debug) printf("MOVE C, OUT1");
			out[1] = c;
			break;
		case 10:
			if (debug) printf("MOVE C, OUT2");
			out[2] = c;
			break;
		case 11:
			if (debug) printf("MOVE C, OUT3");
			out[3] = c;
			break;
		case 12:
		case 13:
			if (debug) printf("<undefined opcode>");
			/* these 2 instructions are undefined */
			break;
		case 14:
			if (debug) printf("NOP");
			break;
		case 15:
			if (debug) printf("MOVE C, IE");
			break;
		}
		break;
	case (7):  /* move (address pointer moves)  */
		switch (operand) {
		case 0:
			if (debug) printf("INC P1");
			p1 = (p1 + 1) & 65535;
			break;
		case 1:
			if (debug) printf("INC P2");
			p2 = (p2 + 1) & 65535;
			break;
		case 2:
			if (debug) printf("DEC P1");
			p1 = (p1 - 1) & 65535;
			break;
		case 3:
			if (debug) printf("DEC P2");
			p2 = (p2 - 1) & 65535;
			break;
		case 4:
			if (debug) printf("MOVE P2, P1");
			p1 = p2;
			break;
		case 5:
			if (debug) printf("MOVE P1, P2");
			p2 = p1;
			break;
		case 6:
			if (debug) printf("MOVE SP, P1");
			p1 = sp;
			break;
		case 7:
			if (debug) printf("MOVE P1, SP");
			sp = p1;
			break;
		case 8:
			if (debug) printf("MOVE P1L, A");
			a = p1 & 255;
			break;
		case 9:
			if (debug) printf("MOVE P1H, A");
			a = (p1 >> 8) & 255;
			break;
		case 10:
			if (debug) printf("MOVE P1L, B");
			b = p1 & 255;
			break;
		case 11:
			if (debug) printf("MOVE P1H, B");
			b = (p1 >> 8) & 255;
			break;
		case 12:
			if (debug) printf("MOVE A, P1L");
			p1 &= 65280;
			p1 |= a & 255;
			break;
		case 13:
			if (debug) printf("MOVE A, P1H");
			p1 &= 255;
			p1 |= (a & 255) << 8;
			break;
		case 14:
			if (debug) printf("MOVE B, P1L");
			p1 &= 65280;
			p1 |= b & 255;
			break;
		case 15:
			if (debug) printf("MOVE B, P1H");
			p1 &= 255;
			p1 |= (b & 255) << 8;
			break;
		}
		if (p1 == p2) p = 1; else p = 0;
		break;
	case (8):
		a &= 240;
		a |= operand;
		if (debug) printf("MOVE #%1x, Ab", operand);
		break;
	case (9):
		a &= 15;
		a |= operand << 4;
		if (debug) printf("MOVE #%1x, At", operand);
		break;
	case (10):
		b &= 240;
		b |= operand;
		if (debug) printf("MOVE #%1x, Bb", operand);
		break;
	case (11):
		b &= 15;
		b |= operand << 4;
		if (debug) printf("MOVE #%1x, Bt", operand);
		break;
	case (12):
		p1 &= 65520;
		p1 |= operand;
		if (debug) printf("MOVE #%1x, P1Lb", operand);
		break;
	case (13):
		p1 &= 65295;
		p1 |= operand << 4;
		if (debug) printf("MOVE #%1x, P1Lt", operand);
		break;
	case (14):
		p1 &= 61695;
		p1 |= operand << 8;
		if (debug) printf("MOVE #%1x, P1Hb", operand);
		break;
	case (15):
		p1 &= 4095;
		p1 |= operand << 12;
		if (debug) printf("MOVE #%1x, P1Ht", operand);
		break;

	default:  /* the alu instructions */
                switch (operand) {
                        case 0:
				alu = a;
				if (debug) printf("MOVE A");
				break;
                        case 1:
				alu = (~a) & 255;
				if (debug) printf("CPL A");
				break;
                        case 2:
				alu = b;
				if (debug) printf("MOVE B");
				break;
                        case 3:
				alu = (~b) & 255;
				if (debug) printf("CPL B");
				break;
                        case 4:
				alu = a + b + c;
				if (opcode == 0) {
					if (alu & 256) c = 1; else c = 0;
					if ( (!(alu & 255) && (a & 255) && (b & 255)) \
					|| ((alu & 255) && !(a & 255) && !(b & 255)) )
						v = 1; else v = 0;
				}
				alu &= 255;
				if (debug) printf("ADD A+B");
				break;
                        case 5:
				alu = a - b - c;
				if (opcode == 0) {
					if (alu & 256) c = 1; else c = 0;
					if ( (!(alu & 255) && (a & 255) && !(b & 255)) \
					|| ((alu & 255) && !(a & 255) && (b & 255)) )
						v = 1; else v = 0;
				}
				alu &= 255;
				if (debug) printf("SUB A-B");
				break;
                        case 6:
				alu = a & b;
				if (debug) printf("AND A&B");
				break;
                        case 7:
				alu = a | b;
				if (debug) printf("OR A|B");
				break;
                        case 8:
				alu = a >> 1;
				alu |= (a & 1) << 7;
				if (debug) printf("RR A");
				break;
                        case 9:
				alu = (a << 1) & 254;
				alu |= (a & 128) >> 7;
				if (debug) printf("RL A");
				break;
                        case 10:
				alu = (a >> 1) & 127;
				alu |= (c & 1) << 7;
				c = a & 1;
				if (debug) printf("RRC A");
				break;
                        case 11:
				alu = (a << 1) & 254;
				alu |= c;
				if (a & 128) c = 1; else c = 0;
				if (debug) printf("RLC A");
				break;
                        case 12:
				alu = a + 1;
				if (opcode == 1) {
					if (alu & 256) c = 1; else c = 0;
				}
				alu &= 255;
				if (debug) printf("INC A");
				break;
                        case 13:
				alu = a - 1;
				if (opcode == 1) {
					if (alu & 256) c = 1; else c = 0;
				}
				alu &= 255;
				if (debug) printf("DEC A");
				break;
                        case 14:
				alu = b + 1;
				/* if (alu & 256) c = 1; else c = 0; */
				alu &= 255;
				if (debug) printf("INC B");
				break;
                        case 15:
				alu = b - 1;
				/* if (alu & 256) c = 1; else c = 0; */
				alu &= 255;
				if (debug) printf("DEC B");
				break;
                }
                if (opcode == 0) a = alu; else b = alu;
		if (opcode == 0) {
                	if (alu == 0) z = 1; else z = 0;
		}
		if (opcode == 0 && operand != 1 && \
		    operand != 12 && operand != 13 && debug)
			printf(", A");
		if (opcode == 1 && operand != 3 && \
		    operand != 14 && operand != 15 && debug)
			printf(", B");
	}
}  /* end of the simulator */

/*****************************************************************/
/* commands */


/* parses a line of intel hex code, stores the data in bytes[] */
/* and the beginning address in addr, and returns a 1 if the */
/* line was valid, or a 0 if an error occured.  The variable */
/* num gets the number of bytes that were stored into bytes[] */

int parse_hex_line(theline, bytes, addr, num, code)
char *theline;
int *addr, *num, *code, bytes[];
{
	int sum, len, cksum;
	char *ptr;
	
	*num = 0;
	if (theline[0] != ':') return 0;
	if (strlen(theline) < 11) return 0;
	ptr = theline+1;
	if (!sscanf(ptr, "%02x", &len)) return 0;
	ptr += 2;
	if ( strlen(theline) < (11 + (len * 2)) ) return 0;
	if (!sscanf(ptr, "%04x", addr)) return 0;
	ptr += 4;
	  /* printf("Line: length=%d Addr=%d\n", len, *addr); */
	if (!sscanf(ptr, "%02x", code)) return 0;
	ptr += 2;
	sum = (len & 255) + ((*addr >> 8) & 255) + (*addr & 255) + (*code & 255);
	while(*num != len) {
		if (!sscanf(ptr, "%02x", &bytes[*num])) return 0;
		ptr += 2;
		sum += bytes[*num] & 255;
		(*num)++;
		if (*num >= 256) return 0;
	}
	if (!sscanf(ptr, "%02x", &cksum)) return 0;
	if ( ((sum & 255) + (cksum & 255)) & 255 ) return 0; /* checksum error */
	return 1;
}

/* loads an intel hex file into the global memory[] array */
/* the command is a string such as "L file.hex" */

void load_file(command)
char *command;
{
	char *filename;
	char line[1000];
	FILE *fin;
	int addr, n, status, bytes[256];
	int i, total=0, lineno=1;
	int minaddr=65536, maxaddr=0;

	filename = command + 1;
	while (isspace(*filename)) filename++;
	printf("Load File: '%s'\n", filename);
	if (strlen(filename) == 0) {
		printf("   Can't load a file without the filename.");
		printf("  '?' for help\n");
		return;
	}
	fin = fopen(filename, "r");
	if (fin == NULL) {
		printf("   Can't open file '%s' for reading.\n", filename);
		return;
	}
	while (!feof(fin) && !ferror(fin)) {
		line[0] = '\0';
		fgets(line, 1000, fin);
		if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
		if (line[strlen(line)-1] == '\r') line[strlen(line)-1] = '\0';
		if (parse_hex_line(line, bytes, &addr, &n, &status)) {
			if (status == 0) {  /* data */
				for(i=0; i<=(n-1); i++) {
					memory[addr] = bytes[i] & 255;
					total++;
					if (addr < minaddr) minaddr = addr;
					if (addr > maxaddr) maxaddr = addr;
					addr++;
				}
			}
			if (status == 1) {  /* end of file */
				fclose(fin);
				printf("   Loaded %d bytes between:", total);
				printf(" %04X to %04X\n", minaddr, maxaddr);
				return;
			}
			if (status == 2) ;  /* begin of file */
		} else {
			printf("   Error: '%s', line: %d\n", filename, lineno);
		}
		lineno++;
	}
}

void print_usage_stats()
{
	int i, j, total=0;

	for (i=0; i<=255; i++) total += usage_stat[i];
	if (total == 0) {
		printf("   No instructions executed yet\n");
		return;
	}
	printf("Opcode Usage Statistics:\n");
	for (i=0; i<=15; i++) {
		for (j=0; j<=15; j++)
			printf("%5.1f", \
			  ((float)usage_stat[j*16+i]/(float)total)*100.0);
		printf("\n");
	}
	printf("Branch Outcomes:\n");
	if (jcu_t+jcu_nt) {
		printf("  JCU taken:     %5.1f%%\n", \
		  ((float)(jcu_t)/(float)(jcu_t+jcu_nt)*100.0));
		printf("      not taken: %5.1f%%\n", \
		  ((float)(jcu_nt)/(float)(jcu_t+jcu_nt)*100.0));
	} else printf("  JCU - none executed yet\n");
	if (jcd_t+jcd_nt) {
		printf("  JCD taken:     %5.1f%%\n", \
		  ((float)(jcd_t)/(float)(jcd_t+jcd_nt)*100.0));
		printf("      not taken: %5.1f%%\n", \
		  ((float)(jcd_nt)/(float)(jcd_t+jcd_nt)*100.0));
	} else printf("  JCD - none executed yet\n");
}

void print_memory(command)
char *command;
{
	static int last_begin=0, last_end=255;
	int begin, end, addr, line_byte[16], num, i;
	char *ptr;

	ptr = command+1;
	while (isspace(*ptr)) ptr++;
	if (*ptr != '\0') {   /* get the parameters */
		if (sscanf(ptr, "%x%x", &begin, &end) < 2) {
			printf("   Invalid addresses, type '?' for help.\n");
			return;
		}
		begin &= 65535;
		end &= 65535;
		if (begin > end) {
			printf("   Begin address must be less than end address.\n");
			return;
		}
		last_begin = begin;
		last_end = end;
	}
	begin = last_begin;
	end = last_end;
	addr = begin;
	while (addr <= end) {
		num = 0;
		printf("%04X:", addr);
		while (addr <= end && num <= 15)
			line_byte[num++] = memory[addr++];
		for (i=0; i <= 15; i++)
			if (i < num) printf(" %02X", line_byte[i]);
			else printf("   ");
		printf("  ");
		for (i=0; i <= 15; i++)
			if (i < num)
				if (isprint((char)line_byte[i])) 
					printf("%c", (char)line_byte[i]);
				else printf(" ");
		printf("\n");
	}
}

void save_file(command)
char *command;
{
	int begin, end, addr;
	char *ptr, filename[200];

	ptr = command+1;
	while (isspace(*ptr)) ptr++;
	if (*ptr == '\0') {
		printf("   Must specify address range and filename\n");
		return;
		}
	if (sscanf(ptr, "%x%x%s", &begin, &end, filename) < 3) {
		printf("   Invalid addresses or filename, '?' for help.\n");
		return;
	}
	begin &= 65535;
	end &= 65535;
	if (begin > end) {
		printf("   Begin address must be less than end address.\n");
		return;
	}
	fhex = fopen(filename, "w");
	if (fhex == NULL) {
		printf("   Can't open '%s' for writing.\n", filename);
		return;
	}
	for (addr=begin; addr <= end; addr++)
		hexout(memory[addr], addr, 0);
	hexout(0, 0, 1);
	printf("Memory %04X to %04X written to '%s'\n", begin, end, filename);
}

void assign_value(command)
char *command;
{
	int value, addr;
	char *ptr, reg_loc[200];

	ptr = command+1;
	while (isspace(*ptr)) ptr++;
	if (*ptr == '\0') {
		printf("   Must specify a value and a register\n");
		return;
		}
	if (sscanf(ptr, "%x%s", &value, reg_loc) < 2) {
		printf("   Invalid input format, '?' for help.\n");
		return;
	}
	if (!strncasecmp(reg_loc, "mem=", 4)) {
		ptr = reg_loc+4;
		if (!sscanf(ptr, "%x", &addr)) {
			printf("   Invalid address, type '?' for help.\n");
			return;
		}
		addr &= 65535;
		value &= 255;
		memory[addr] = value;
		printf("Wrote %02X to memory location %04X\n", value, addr);
		return;
	}
	if (!strcasecmp(reg_loc, "a")) {a = value & 255; reg();}
	else if (!strcasecmp(reg_loc, "b")) {b = value & 255; reg();}
	else if (!strcasecmp(reg_loc, "p1")) {p1 = value & 65535; reg();}
	else if (!strcasecmp(reg_loc, "p2")) {p2 = value & 65535; reg();}
	else if (!strcasecmp(reg_loc, "sp")) {sp = value & 65535; reg();}
	else if (!strcasecmp(reg_loc, "pc")) {pc = value & 65535; reg();}
	else if (!strcasecmp(reg_loc, "c")) {c = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "v")) {v = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "z")) {z = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "p")) {p = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "in0")) {in[0] = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "in1")) {in[1] = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "out0")) {out[0] = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "out1")) {out[1] = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "out2")) {out[2] = value & 1; reg();}
	else if (!strcasecmp(reg_loc, "out3")) {out[3] = value & 1; reg();}
	else {
		printf("  Valid register names are: A, B, P1, P2, SP, PC, MEM=\n");
		printf("    C, V, Z, P, IN0, IN1, OUT0, OUT1, OUT2, OUT3\n");
	}
}

void help()
{
	printf("OSU-8 Simulator, Paul Stoffregen, March 1995\n\n");
	printf("   <return>    - Execute next instruction.\n");
	printf("  A value reg  - Assign a value to a register.  'value' in hex\n");
	printf("                 'reg' is a register name or 'mem=address' to\n");
	printf("                 assign the value to a memory location (hex).\n");
	printf("  D            - Toggle disassembling and printing of opcodes.\n");
	printf("  Debug        - Toggle disassember debugging mode.\n");
	printf("  E number     - Execute next 'number' of instructions with no\n");
	printf("                 prompting.  'number' is decimal (base10)\n");
	printf("  L filename   - Load an intel hex file into memory\n");
	printf("  M begin end  - Print the contents of memory from 'begin' to\n");
	printf("                 'end', which must be hexidecimal numbers\n");
	printf("  M            - With no parameters, prints same range again\n");
	printf("  R            - Toggle printing of register values\n");
	printf("  Reset        - Reset the processor, memory is unchanged\n");
	printf("  S begin end filename\n");
	printf("                 Save memory from 'begin' to 'end' in filename\n");
	printf("                 using intel hex format, begin/end must be hex.\n");
	printf("  Usage        - Print a cryptic instruction usage summary\n");
        printf("  Q            - Quit\n");
}

main(argc, argv)
int argc;
char *argv[];
{
	int	i, j;
	long	ck1, ck2;
	char	inbuf[1000], *pchar;
	int	dodisasm=1, doreg=1;

	if (argc > 1) printf("Command line parameters ignored, type '?' for help\n");
	printf("\nOSU-8 Instructional Microprocessor Simulator\n");
	printf("  by Paul Stoffregen, March 1995\n");
	printf("\nType '?' for help\n\n");
	/* Let's be nice and clear everything to zero */
	/* though in the real chip, only the PC needs to */
	/* be zero on startup */
	a = b = p1 = p2 = sp = pc = 0;
	c = v = z = p = 0;
	in[0] = in[1] = out[0] = out[1] = out[2] = out[3] = 0;
	for(i=0; i<65536; i++) memory[i] = 0;
	for(i=0; i<256; i++) usage_stat[i] = 0;
	reg();
	while (!feof(stdin) && !ferror(stdin)) {
		inbuf[0] = '\0';
		fgets(inbuf, 1000, stdin);
		if (inbuf[strlen(inbuf)-1] == '\n') inbuf[strlen(inbuf)-1] = '\0';
		if (inbuf[strlen(inbuf)-1] == '\r') inbuf[strlen(inbuf)-1] = '\0';
		/* now let's do something */
		if (inbuf[0] == '\0') {
			if (dodisasm) printf("  %s  ", disasm(memory[pc], pc));
			one_instruction();
			if (dodisasm) printf("\n");
			if (doreg) reg();
			continue;
		}

		if (!strncasecmp(inbuf, "usage", 5)) {
			print_usage_stats();
			continue;
		}
		if (!strncasecmp(inbuf, "debug", 5)) {
			if (debug) {
				printf("Disassembler debugging turned off.\n");
				debug = 0;
			} else {
				printf("Disassembler debugging turned on.\n");
				debug = 1;
			}
			continue;
		}
		if (!strncasecmp(inbuf, "reset", 5)) {
			a = b = p1 = p2 = sp = pc = 0;
			c = v = z = p = 0;
			out[0] = out[1] = out[2] = out[3] = 0;
			printf("OSU-8 Microprocessor RESET\n");
			reg();
			continue;
		}

		if (inbuf[0] == '?') {
			help();
			continue;
		}
		if (inbuf[0] == 'q' || inbuf[0] == 'Q') break;
		if (inbuf[0] == 'l' || inbuf[0] == 'L') {
			load_file(inbuf);
			continue;
		}
		if (inbuf[0] == 'd' || inbuf[0] == 'D') {
			if (dodisasm) {
				printf("Disassembling turned off.\n");
				dodisasm = 0;
				debug = 0;
			} else {
				printf("Disassembling turned on.\n");
				dodisasm = 1;
			}
			continue;
		}
		if (inbuf[0] == 'r' || inbuf[0] == 'R') {
			if (doreg) {
				printf("Register printing turned off.\n");
				doreg = 0;
			} else {
				printf("Register printing turned on.\n");
				doreg = 1;
				reg();
			}
			continue;
		}
		if ((inbuf[0] == 'e' || inbuf[0] == 'E') && inbuf[1] == '\0') {
			printf("   Must specify number of instructions to execute.\n");
			continue;
		}
		if ((inbuf[0] == 'e' || inbuf[0] == 'E') && isspace(inbuf[1])) {
			pchar = inbuf+1;
			while (isspace(*pchar)) pchar++;
			if ((*pchar != '\0') && sscanf(pchar, "%d", &i)) {
				printf("Execute %d instrustions:\n", i);
				ck1 = (long)clock();
				for(j=1; j<=i; j++) {
					if (dodisasm) printf("  %s  ", disasm(memory[pc], pc));
					one_instruction();
					if (dodisasm) printf("\n");
					if (doreg) reg();
				}
				ck2 = (long)clock();
				printf("Finished %d instructions, cpu time: %6.3f sec\n", i, \
				   ((float)(ck2 - ck1) / (float)(CLOCKS_PER_SEC)) );
					/* is "CLOCKS_PER_SEC" undefined, try "CLK_TCK" */
			} else {
				printf("Execute how many instructions??  Type '?' for help\n");
			}
			continue;
		}
		if (inbuf[0] == 'm' || inbuf[0] == 'M') {
			print_memory(inbuf);
			continue;
		}
		if (inbuf[0] == 's' || inbuf[0] == 'S') {
			save_file(inbuf);
			continue;
		}
		if (inbuf[0] == 'a' || inbuf[0] == 'A') {
			assign_value(inbuf);
			continue;
		}

		printf("   Unrecognized command '%s'.  Type '?' for help.\n", inbuf);
	}
}


